from importlib import import_module
import os

from service import Service
from ..hooks import hook
from ..shared import bold, red

def get_aws_access_key(site):
	if site.has_key('webapp') and site['webapp'].get('aws_access_key_id') and site['webapp'].get('aws_secret_access_key'):
		return site['webapp']['aws_access_key_id'], site['webapp']['aws_secret_access_key']
	else:
		try:
			module = import_module(site['settings_module'])
			return module.AWS_ACCESS_KEY_ID, module.AWS_SECRET_ACCESS_KEY
		except:
			print red('Error: Could not access AWS_ACCESS_KEY_ID or AWS_SECRET_ACCESS_KEY from settings.')
			exit(1)

def setup_aws_access_key(site):
	# Boto variables
	aws_access_key_id, aws_secret_access_key = get_aws_access_key(site)
	os.environ.setdefault('AWS_ACCESS_KEY_ID', aws_access_key_id)
	os.environ.setdefault('AWS_SECRET_ACCESS_KEY', aws_secret_access_key)

def upload_to_s3(site, bucket, directory=None, files=None, prefix=None):
	"""Uploads files to an s3 bucket. Upload either an entire directory with files=None, or specific files with and optional directory prefix."""
	if bucket is None:
		print red('Error: Bucket must be specified.')
		return
	if directory is None and files is None:
		print red('Error: Directory and/or files must be specified.')
		return
	# Setup boto
	import boto
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key
	import mimetypes
	import fnmatch

	setup_aws_access_key(site)

	# Connect to S3
	c = boto.connect_s3()
	b = Bucket(c, bucket)

	# Fix the prefix
	# prefix itself shouldn't have a / prefix itself but should end with /
	if prefix:
		prefix = prefix.lstrip('/')
		if prefix and not prefix.endswith('/'):
			prefix = prefix + '/'

	def __upload(key, filename):
		k = Key(b)
		k.key = key
		headers = {}
		content_type = mimetypes.guess_type(filename)[0]
		if site.has_key('webapp') and site['webapp'].get('cache_control'):
			for pattern in site['webapp']['cache_control']:
				if fnmatch.fnmatch(filename, pattern):
					headers['Cache-Control'] = site['webapp']['cache_control'][pattern]
					break
		if site.has_key('webapp') and site['webapp'].get('gzip_types') and content_type in site['webapp']['gzip_types']:
			from gzip import GzipFile
			from StringIO import StringIO
			# Need to specify content_type when uploading from a string!
			headers['Content-Type'] = content_type
			headers['Content-Encoding'] = 'gzip'
			s = StringIO()
			g = GzipFile(fileobj=s, mode='wb')
			with open(filename, 'rb') as f:
				g.write(f.read())
			g.close()
			k.set_contents_from_string(s.getvalue(), headers)
		else:
			k.set_contents_from_filename(filename, headers)

	if files:
		# Upload individual files
		if directory:
			keys = [filename.lstrip('/') for filename in files]
			files = [os.path.join(directory, filename) for filename in files]
		else:
			keys = [os.path.split(filename)[1] for filename in files]
		for i, filename in enumerate(files):
			print 'Uploading %s' % keys[i]
			if prefix:
				key = prefix + keys[i]
			else:
				key = keys[i]
			__upload(key, filename)
	elif directory:
		# Upload an entire directory
		def __upload_dir(arg, dirname, names):
			# arg is the starting directory
			for name in names:
				filename = os.path.join(dirname, name)
				if not os.path.isdir(filename) and not os.path.islink(filename) and not name.startswith('.'):
					key = filename[len(arg):]
					if key.startswith('/'):
						key = key[1:]
					if prefix:
						key = prefix + key
					print 'Uploading %s' % key
					__upload(key, filename)
		os.path.walk(directory, __upload_dir, directory)

def delete_from_s3(site, bucket, prefix=None):
	""" Remove all files with the prefix specified from the bucket. """
	if bucket is None:
		print red('Error: Bucket must be specified.')
		return
	# Setup boto
	import boto
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key

	setup_aws_access_key(site)

	# Fix the prefix
	if prefix:
		prefix = prefix.lstrip('/')

	# Connect to S3, list the contents, and remove all of the keys
	c = boto.connect_s3()
	b = Bucket(c, bucket)
	result_set = b.list(prefix=prefix)
	result = b.delete_keys([key.name for key in result_set])

def create_public_bucket_policy(bucket_name):
	import json
	policy = {
		"Version": "2008-10-17",
		"Statement": [
			{
				"Sid": "AddPerm",
				"Effect": "Allow",
				"Principal": {
					"AWS": "*"
				},
				"Action": "s3:GetObject",
				"Resource": "arn:aws:s3:::%s/*" % bucket_name
			}
		]
	}
	return json.dumps(policy)

DEFAULT_CORS_RULE = {'allowed_method': ['GET'], 'allowed_origin': ['*'], 'allowed_header': ['Authorization'], 'max_age_seconds': 3000}

class S3(Service):
	name = 's3'
	script = None

	def site_config(self, site):
		with hook('site config %s' % self.name, self, site):
			setup_aws_access_key(site)

			from boto import connect_s3
			from boto.s3.bucket import Bucket
			from boto.s3.key import Key

			for bucket_config in self.settings['buckets']:
				# Connect and make sure the bucket exists
				print bold(u'Configuring bucket %s...' % bucket_config['name'])
				connection = connect_s3()
				try:
					bucket = connection.get_bucket(bucket_config['name'])
				except:
					bucket = connection.create_bucket(bucket_config['name'])
				# Set the bucket policy
				if bucket_config.has_key('policy'):
					bucket.set_policy(bucket_config['policy'])
				# Setup CORS, array of rules
				# http://boto.readthedocs.org/en/latest/ref/s3.html#boto.s3.cors.CORSConfiguration
				if bucket_config.has_key('cors') and bucket_config['cors'] is None:
					# If explicity set to None, then remove the cors policy
					bucket.delete_cors()
				else:
					if not bucket_config.has_key('cors'):
						# If not specified, use the default GET policy
						bucket_config['cors'] = (DEFAULT_CORS_RULE,)
					from boto.s3.cors import CORSConfiguration
					cors_config = CORSConfiguration()
					for rule in bucket_config['cors']:
						cors_config.add_rule(**rule)
					bucket.set_cors(cors_config)
				# Setup the lifecycle, array of rules
				# http://boto.readthedocs.org/en/latest/ref/s3.html#boto.s3.lifecycle.Lifecycle
				if bucket_config.has_key('lifecycle'):
					from boto.s3.lifecycle import Lifecycle
					lifecycle_config = Lifecycle()
					for rule in bucket_config['lifecycle']:
						lifecycle_config.add_rule(**rule)
					bucket.configure_lifecycle(lifecycle_config)
				else:
					bucket.delete_lifecycle_configuration()
				# Setup the bucket website hosting {suffix, error_key, routing_rules, redirect_all_requests_to}
				# http://boto.readthedocs.org/en/latest/ref/s3.html
				# https://github.com/boto/boto/blob/develop/boto/s3/website.py
				if bucket_config.has_key('website'):
					# Expand the routing rules, array of {condition, redirect}
					if bucket_config['website'].has_key('routing_rules'):
						from boto.s3.website import RoutingRules, RoutingRule
						routing_rules = RoutingRules()
						for rule in bucket_config['website']['routing_rules']:
							routing_rules.add_rule(RoutingRule(**rule))
						bucket_config['website']['routing_rules'] = routing_rules
					# Expand the redirect, redirect_all_requests_to is {hostname, protocol}
					if bucket_config['website'].has_key('redirect_all_requests_to'):
						from boto.s3.website import RedirectLocation
						bucket_config['website']['redirect_all_requests_to'] = RedirectLocation(**bucket_config['website']['redirect_all_requests_to'])
					bucket.configure_website(**bucket_config['website'])
				else:
					bucket.delete_website_configuration()



from fabric.api import cd, sudo
from fabric.contrib.files import exists

from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, pip_install, SiteType

# http://dev.maxmind.com/geoip/legacy/geolite/
_CITY_PACKAGE_URL = 'http://geolite.maxmind.com/download/geoip/database/GeoLiteCity.dat.gz'
_COUNTRY_PACKAGE_URL = 'http://geolite.maxmind.com/download/geoip/database/GeoLiteCountry/GeoIP.dat.gz'
_INSTALL_DIR = '/opt/geoip'

class GeoIP(Service):
	"""Database to convert ip addresses into locations."""
	name = 'geoip'
	script = None

	def install(self):
		with hook('install %s' % self.name, self):
			if not exists(_INSTALL_DIR):
				apt_get_install('libgeoip1')
				sudo('mkdir %s' % _INSTALL_DIR)
				with cd(_INSTALL_DIR):
					sudo('wget --no-clobber %s' % _CITY_PACKAGE_URL)
					sudo('wget --no-clobber %s' % _COUNTRY_PACKAGE_URL)
					sudo('gunzip %s' % _CITY_PACKAGE_URL.split('/')[-1])
					sudo('gunzip %s' % _COUNTRY_PACKAGE_URL.split('/')[-1])

	def get_site_settings(self, site):
		if site['type'] == SiteType.DJANGO:
			return { 'GEOIP_PATH': _INSTALL_DIR }
		else:
			return super(GeoIP, self).get_site_settings(site)



from setuptools import setup, find_packages

setup(
	name='fabric-shuttle',
	version='0.0.6',
	description='Declarative service-oriented approach with hooks for using Fabric to provision a machine or vagrant box.',
	url='https://github.com/mvx24/fabric-shuttle',
	author='mvx24',
	author_email='cram2400@gmail.com',
	license='MIT',
	classifiers=[
		'Development Status :: 4 - Beta',
		'Environment :: Console',
		'Framework :: Django',
		'License :: OSI Approved :: MIT License',
		'Operating System :: MacOS :: MacOS X',
		'Operating System :: POSIX',
		'Operating System :: Unix',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 2 :: Only',
		'Topic :: System :: Installation/Setup',
		'Topic :: System :: Software Distribution',
		'Topic :: System :: Systems Administration'
	],
	keywords='fabric deployment development vagrant provisioning services',
	packages=find_packages(),
	install_requires=['fabric', 'boto', 'django', 'pyyaml'],
	package_data={
		'': ['templates/*']
	}
)



from ..shared import bold, green

_before_hooks = {}
_after_hooks = {}
_hook_depth = 0
_quiet_hooks = False

def before(hook_name):
	def before_decorator(func):
		if not _before_hooks.has_key(hook_name):
			_before_hooks[hook_name] = [func]
		else:
			_before_hooks[hook_name].append(func)
		return func
	return before_decorator

def after(hook_name):
	def after_decorator(func):
		if not _after_hooks.has_key(hook_name):
			_after_hooks[hook_name] = [func]
		else:
			_after_hooks[hook_name].append(func)
		return func
	return after_decorator

class hook(object):
	""" Run registered the hooks. """
	def __init__(self, hook_name, *args, **kwargs):
		self.hook_name = hook_name
		self.args = args
		self.kwargs = kwargs
	def __enter__(self):
		global _hook_depth
		if not _quiet_hooks:
			print bold('%s>Start %s' % ('-' * _hook_depth, self.hook_name))
		if _before_hooks.has_key(self.hook_name):
			for func in _before_hooks[self.hook_name]:
				if func.__doc__:
					print bold(func.__doc__)
				func(*self.args, **self.kwargs)
		_hook_depth += 1
	def __exit__(self, *_):
		global _hook_depth
		_hook_depth -= 1
		if _after_hooks.has_key(self.hook_name):
			for func in _after_hooks[self.hook_name]:
				if func.__doc__:
					print bold(func.__doc__)
				func(*self.args, **self.kwargs)
		if not _quiet_hooks:
			print green('%s>End %s' % ('-' * _hook_depth, self.hook_name))

class quiet_hook_messages(object):
	"""Silence hook output down to one message."""
	def __init__(self, msg):
		print green(msg)
	def __enter__(self):
		global _quiet_hooks
		_quiet_hooks = True
	def __exit__(self, *_):
		global _quiet_hooks
		_quiet_hooks = False
		print bold('Done.')



import tempfile

from fabric.api import put, sudo

from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, pip_install

class MySQL(Service):
	name = 'mysql'
	script = 'mysql'

	def install(self):
		with hook('install %s' % self.name, self):
			apt_get_install('mysql-server')

	def config(self):
		with hook('config %s' % self.name, self):
			if self.settings:
				with tempfile.NamedTemporaryFile('w') as f:
					for section in self.settings:
						f.write('[%s]\n' % section)
						for setting in self.settings[section]:
							f.write('%s = %s\n' % (setting, self.settings[section][setting]))
					f.flush()
					put(f.name, '/etc/mysql/conf.d/fabric.cnf', use_sudo=True, mode=0644)		
					sudo('chown root:root /etc/mysql/conf.d/fabric.cnf')
		self.restart()

	def site_install(self, site):
		with hook('site install %s' % self.name, self, site):
			apt_get_install('mysql-client', 'libmysqlclient-dev')
			pip_install('mysql-python')



from importlib import import_module
import os
import subprocess

from fabric.api import sudo, local, put, settings
from fabric.contrib.files import append, upload_template

from .services.s3 import upload_to_s3, delete_from_s3
from .shared import *

class own_project(object):
	"""Changes the permissions of the project between the ssh user and nginx user."""
	def __init__(self):
		pass
	def __enter__(self):
		chown(get_project_directory(), env['user'], env['user'])
	def __exit__(self, *_):
		chown(get_project_directory(), WWW_USER, WWW_USER)

def _get_remote_shell():
	parts = ['ssh', '-p', env.get('port', '22')]
	if env.get('key_filename'):
		parts.append('-i')
		parts.append(os.path.expanduser(env['key_filename']))
	if env.get('vagrant'):
		parts.append('-o')
		parts.append('UserKnownHostsFile=/dev/null')
		parts.append('-o')
		parts.append('StrictHostKeyChecking=no')
	return ' '.join(parts)

def _django_get_excluded(sites):
	"""Get an excluded list of patterns.

	rsync patterns are similar to .gitignore patterns. e.g. /foo matches foo at the base.
	http://git-scm.com/docs/gitignore
	http://ss64.com/bash/rsync.html
	"""
	excluded = ['.*', '*.pyc', '*.db']
	# Exclude anything additionally specified in the environment
	if env.has_key('excluded_files') and env['excluded_files']:
		excluded.extend(env['excluded_files'])
	# Exclude anything in the root .gitignore file
	if os.path.exists('.gitignore'):
		with file('.gitignore', 'r') as f:
			for line in f:
				line = line.strip()
				if line.startswith('.') or line.find('!') != -1:
					continue
				if line not in excluded:
					excluded.append(line)
	# Exlude all of the webapp's node_modules and bower_components
	for site in sites:
		module = import_module(site['settings_module'])
		os.environ.setdefault('DJANGO_SETTINGS_MODULE', site['settings_module'])
		if hasattr(module, 'WEBAPP_ROOT'):
			webapp_root = module.WEBAPP_ROOT
			try:
				parent, task_runner = get_webapp_taskrunner(webapp_root)
				if task_runner:
					excluded.append('/' + os.path.relpath(parent) + '/node_modules')
					excluded.append('/' + os.path.relpath(parent) + '/bower_components')
				# If wanting to exclude the entire webapp
				#if task_runner:
				#	excluded.append('/' + os.path.relpath(parent))
				#else:
				#	excluded.append('/' + os.path.relpath(webapp_root))
			except:
				print red('Warning: Could not import webapp settings for %s when syncing.' % site['name'])
	return excluded

SYNC_COMMAND = 'rsync -avz %s --delete --delete-excluded%s ./ %s@%s:%s'

def django_sync(sites):
	"""Syncs the local code to the server. Used as part of the deploy process. Wrap in the task decorator to enable. sync = task(sync)"""
	project_dir = project_subpath = get_project_directory()
	if not project_subpath.endswith('/'):
		project_subpath += '/'
	sudo('mkdir -p %s' % project_dir)
	manage_dir = get_manage_directory()
	sudo('mkdir -p %s' % manage_dir)
	with own_project():
		# Preserve existing media if a subdirectory of the project
		for site in sites:
			media_root = get_media_root(site)
			if media_root.startswith(project_subpath) and exists(media_root):
				sudo('mv %s /tmp/%smedia' % (media_root, env['project']))

		# Sync and set the owner
		excluded = _django_get_excluded(sites)
		excluded = ['--exclude="%s"' % ex for ex in excluded]
		local(SYNC_COMMAND % (' '.join(excluded), ' -e "%s"' % _get_remote_shell(), env['user'], env['hosts'][0], project_dir))

		# Create the media directory
		# Restore existing media if a subdirectory of the project, if it is outside the project make sure it is owned by nginx
		for site in sites:
			media_root = get_media_root(site)
			sudo('mkdir -p %s' % media_root)
			if media_root.startswith(project_subpath):
				if exists('/tmp/%smedia' % env['project']):
					# Delete the media directory but leave intermediate directories then restore with a move
					sudo('rm -rf %s' % media_root)
					sudo('mv /tmp/%smedia %s' % (env['project'], media_root))
			else:
				chown(media_root, WWW_USER, WWW_USER)

			# Copy any additional webapp files
			if site.has_key('webapp') and site['webapp'].get('files'):
				from django.contrib.staticfiles import finders
				webapp_root = get_webapp_root(site)
				if webapp_root:
					for filename in site['webapp']['files']:
						result = finders.find(filename)
						if not result:
							print red('Error: Could not find static file %s.' % filename)
							return
						result = result[0] if isinstance(result, (list, tuple)) else result
						chown(put(result, os.path.join(webapp_root, filename), use_sudo=True, mode=0644), WWW_USER, WWW_USER)

		for site in sites:
			# Create manage.py shortcuts the manage directory
			context = {'project_dir': project_dir, 'settings_module': site['settings_module'], 'interpreter': get_python_interpreter(site)}
			chown(upload_template(get_template('manage.py'), os.path.join(manage_dir, site['name'] + '.py'), context=context, backup=False, use_sudo=True, mode=0755), WWW_USER, WWW_USER)

def django_sync_dry_run(sites):
	"""Do an rsync dry run to see which files will be updated when deploying."""
	# e.g. To show just migrations: fab e:production x:dryrun | grep -v "^deleting" | grep -v "/$" | grep "^shared/migrations"
	excluded = _django_get_excluded(sites)
	excluded = ['--exclude="%s"' % ex for ex in excluded]
	local(SYNC_COMMAND % (' '.join(excluded), ' --dry-run -e "%s"' % _get_remote_shell(), env['user'], env['hosts'][0], get_project_directory()))

def django_sync_down(path=''):
	"""Syncs down stuff from the server. Wrap in the task decorator to enable. sync_down = task(sync_down)"""
	if path.startswith('/'):
		path = path[1:]
	if not path.endswith('/'):
		path += '/'
	remote_path = os.path.join(get_project_directory(), path)
	local('rsync -avz --exclude=".*" --exclude="*.pyc" --exclude="*.sh" --exclude="*.db" -e "%s" %s@%s:%s ./%s' % (_get_remote_shell(), env['user'], env['hosts'][0], remote_path, path))

def django_append_settings(site):
	"""Update Django settings with production ready values, possibly not set explicitly in the settings."""
	filename = fix_absolute_path(os.path.abspath(site['settings_module'].replace('.', '/'))) + '.py'
	txt = """
	####################################
	# Default/Fixed Paths from Shuttle #

	MEDIA_ROOT = '%s'
	MEDIA_URL = '%s'
	STATIC_ROOT = '%s'
	STATIC_URL = '%s'
	""" % (get_media_root(site), get_media_url(site), get_static_root(site), get_static_url(site))
	webapp_root = get_webapp_root(site)
	if webapp_root:
		txt += "\nWEBAPP_ROOT = '%s'\nWEBAPP_URL = '%s'\n" % (webapp_root, get_webapp_url(site))
	# For vagrant dev boxes add localhost to the ALLOWED_HOSTS
	if env.get('vagrant'):
		allowed_hosts = ', '.join(["'%s'" % host for host in (get_django_setting(site, 'ALLOWED_HOSTS') or [site['name']])])
		txt += "\nALLOWED_HOSTS = [%s, 'localhost']\n" % allowed_hosts
	# Each service might have additional settings
	for service in site.get('services', []):
		site_settings = service.get_site_settings(site)
		if site_settings:
			txt += '\n' + '\n'.join(["%s = '%s'" % item for item in site_settings.items()]) + '\n'
	append(filename, txt.replace('\t', ''))

def deploy_webapp():
	"""Deploy a webapp to S3 with the prefix of WEBAPP_URL. If site is not specified, then the command will be run on all sites."""
	site = env.get('site')
	if not site:
		for site in env['sites'].values():
			env['site'] = site
			deploy_webapp()
		del env['site']
		return

	if site['type'] == SiteType.DJANGO:
		root = get_django_setting(site, 'WEBAPP_ROOT')
		if not root:
			return
		prefix = get_webapp_url(site)
	else:
		if not site.has_key('webapp'):
			return
		root = site['webapp']['root']
		prefix = None

	# Build the webapp if needed
	task = site['webapp']['build_task'] if site.has_key('webapp') and site['webapp'].has_key('build_task') else None
	build_webapp(root, task)
	# Don't go any further if there is no bucket specified
	if not site.has_key('webapp') or not site['webapp'].has_key('bucket'):
		return
	# Clear the bucket
	if site['webapp'].get('clear'):
		delete_from_s3(site, site['webapp']['bucket'], prefix=prefix)
	# Upload the webapp files
	upload_to_s3(site, site['webapp']['bucket'], root, prefix=prefix)
	# Upload the static files found in other directories with the static file finder
	if site['webapp'].get('files'):
		if site['type'] == SiteType.DJANGO:
			from django.contrib.staticfiles import finders
			for filename in site['webapp']['files']:
				result = finders.find(filename)
				if not result:
					print red('Error: Could not find static file %s.' % filename)
					return
				result = result[0] if isinstance(result, (list, tuple)) else result
				upload_to_s3(site, site['webapp']['bucket'], result[:-len(filename)], (filename,), prefix=prefix)
		else:
			if type(site['webapp']['files']) is dict:
				for prefix in site['webapp']['files']:
					upload_to_s3(site, site['webapp']['bucket'], None, site['webapp']['files'][prefix], prefix=prefix)
			else:
				upload_to_s3(site, site['webapp']['bucket'], None, site['webapp']['files'], prefix=prefix)
